"""Test the /files/ handler."""

import io
import os
from unicodedata import normalize

pjoin = os.path.join

import requests
import json

from nbformat import write
from nbformat.v4 import (new_notebook,
                              new_markdown_cell, new_code_cell,
                              new_output)

from notebook.utils import url_path_join
from .launchnotebook import NotebookTestBase
from ipython_genutils import py3compat


class FilesTest(NotebookTestBase):
    def test_hidden_files(self):
        not_hidden = [
            u'å b',
            u'å b/ç. d',
        ]
        hidden = [
            u'.å b',
            u'å b/.ç d',
        ]
        dirs = not_hidden + hidden
        
        nbdir = self.notebook_dir
        for d in dirs:
            path = pjoin(nbdir, d.replace('/', os.sep))
            if not os.path.exists(path):
                os.mkdir(path)
            with open(pjoin(path, 'foo'), 'w') as f:
                f.write('foo')
            with open(pjoin(path, '.foo'), 'w') as f:
                f.write('.foo')

        for d in not_hidden:
            path = pjoin(nbdir, d.replace('/', os.sep))
            r = self.request('GET', url_path_join('files', d, 'foo'))
            r.raise_for_status()
            self.assertEqual(r.text, 'foo')
            r = self.request('GET', url_path_join('files', d, '.foo'))
            self.assertEqual(r.status_code, 404)

        for d in hidden:
            path = pjoin(nbdir, d.replace('/', os.sep))
            for foo in ('foo', '.foo'):
                r = self.request('GET', url_path_join('files', d, foo))
                self.assertEqual(r.status_code, 404)

        self.notebook.contents_manager.allow_hidden = True
        try:
            for d in not_hidden:
                path = pjoin(nbdir, d.replace('/', os.sep))
                r = self.request('GET', url_path_join('files', d, 'foo'))
                r.raise_for_status()
                self.assertEqual(r.text, 'foo')
                r = self.request('GET', url_path_join('files', d, '.foo'))
                r.raise_for_status()
                self.assertEqual(r.text, '.foo')

            for d in hidden:
                path = pjoin(nbdir, d.replace('/', os.sep))
                for foo in ('foo', '.foo'):
                    r = self.request('GET', url_path_join('files', d, foo))
                    r.raise_for_status()
                    self.assertEqual(r.text, foo)
        finally:
            self.notebook.contents_manager.allow_hidden = False

    def test_contents_manager(self):
        "make sure ContentsManager returns right files (ipynb, bin, txt)."

        nbdir = self.notebook_dir

        nb = new_notebook(
            cells=[
                new_markdown_cell(u'Created by test ³'),
                new_code_cell("print(2*6)", outputs=[
                    new_output("stream", text="12"),
                ])
            ]
        )

        with io.open(pjoin(nbdir, 'testnb.ipynb'), 'w', 
            encoding='utf-8') as f:
            write(nb, f, version=4)

        with io.open(pjoin(nbdir, 'test.bin'), 'wb') as f:
            f.write(b'\xff' + os.urandom(5))
            f.close()

        with io.open(pjoin(nbdir, 'test.txt'), 'w') as f:
            f.write(u'foobar')
            f.close()

        r = self.request('GET', 'files/testnb.ipynb')
        self.assertEqual(r.status_code, 200)
        self.assertIn('print(2*6)', r.text)
        json.loads(r.text)

        r = self.request('GET', 'files/test.bin')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.headers['content-type'], 'application/octet-stream')
        self.assertEqual(r.content[:1], b'\xff')
        self.assertEqual(len(r.content), 6)

        r = self.request('GET', 'files/test.txt')
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.headers['content-type'], 'text/plain; charset=UTF-8')
        self.assertEqual(r.text, 'foobar')
    
    def test_download(self):
        nbdir = self.notebook_dir
        
        text = 'hello'
        with open(pjoin(nbdir, 'test.txt'), 'w') as f:
            f.write(text)
        
        r = self.request('GET', 'files/test.txt')
        disposition = r.headers.get('Content-Disposition', '')
        self.assertNotIn('attachment', disposition)

        r = self.request('GET', 'files/test.txt?download=1')
        disposition = r.headers.get('Content-Disposition', '')
        self.assertIn('attachment', disposition)
        self.assertIn("filename*=utf-8''test.txt", disposition)
        
    def test_view_html(self):
        nbdir = self.notebook_dir
        
        html = '<div>Test test</div>'
        with open(pjoin(nbdir, 'test.html'), 'w') as f:
            f.write(html)
        
        r = self.request('GET', 'view/test.html')
        self.assertEqual(r.status_code, 200)

    def test_old_files_redirect(self):
        """pre-2.0 'files/' prefixed links are properly redirected"""
        nbdir = self.notebook_dir
        
        os.mkdir(pjoin(nbdir, 'files'))
        os.makedirs(pjoin(nbdir, 'sub', 'files'))
        
        for prefix in ('', 'sub'):
            with open(pjoin(nbdir, prefix, 'files', 'f1.txt'), 'w') as f:
                f.write(prefix + '/files/f1')
            with open(pjoin(nbdir, prefix, 'files', 'f2.txt'), 'w') as f:
                f.write(prefix + '/files/f2')
            with open(pjoin(nbdir, prefix, 'f2.txt'), 'w') as f:
                f.write(prefix + '/f2')
            with open(pjoin(nbdir, prefix, 'f3.txt'), 'w') as f:
                f.write(prefix + '/f3')

            url = url_path_join('notebooks', prefix, 'files', 'f1.txt')
            r = self.request('GET', url)
            self.assertEqual(r.status_code, 200)
            self.assertEqual(r.text, prefix + '/files/f1')

            url = url_path_join('notebooks', prefix, 'files', 'f2.txt')
            r = self.request('GET', url)
            self.assertEqual(r.status_code, 200)
            self.assertEqual(r.text, prefix + '/files/f2')

            url = url_path_join('notebooks', prefix, 'files', 'f3.txt')
            r = self.request('GET', url)
            self.assertEqual(r.status_code, 200)
            self.assertEqual(r.text, prefix + '/f3')



"""
Test the notification area and widgets
"""
import pytest

from .utils import wait_for_selector, wait_for_script_to_return_true


def get_widget(notebook, name):
    return notebook.browser.execute_script(
        f"return IPython.notification_area.get_widget('{name}') !== undefined"
    )


def widget(notebook, name):
    return notebook.browser.execute_script(
        f"return IPython.notification_area.widget('{name}') !== undefined"
    )


def new_notification_widget(notebook, name):
    return notebook.browser.execute_script(
        f"return IPython.notification_area.new_notification_widget('{name}') !== undefined"
    )


def widget_has_class(notebook, name, class_name):
    return notebook.browser.execute_script(
        f"""
        var w = IPython.notification_area.get_widget('{name}');
        return w.element.hasClass('{class_name}');
        """
    )


def widget_message(notebook, name):
    return notebook.browser.execute_script(
        f"""
        var w = IPython.notification_area.get_widget('{name}');
        return w.get_message();
        """
    )


def test_notification(notebook):
    # check that existing widgets are there
    assert get_widget(notebook, "kernel") and widget(notebook, "kernel"),\
        "The kernel notification widget exists"
    assert get_widget(notebook, "notebook") and widget(notebook, "notebook"),\
        "The notebook notification widget exists"

    # try getting a non-existent widget
    with pytest.raises(Exception):
        get_widget(notebook, "foo")

    # try creating a non-existent widget
    assert widget(notebook, "bar"), "widget: new widget is created"

    # try creating a widget that already exists
    with pytest.raises(Exception):
        new_notification_widget(notebook, "kernel")

    # test creating 'info', 'warning' and 'danger' messages
    for level in ("info", "warning", "danger"):
        notebook.browser.execute_script(f"""
            var tnw = IPython.notification_area.widget('test');
            tnw.{level}('test {level}');
        """)
        wait_for_selector(notebook.browser, "#notification_test", visible=True)

        assert widget_has_class(notebook, "test", level), f"{level}: class is correct"
        assert widget_message(notebook, "test") == f"test {level}", f"{level}: message is correct"

    # test message timeout
    notebook.browser.execute_script("""
        var tnw = IPython.notification_area.widget('test');
        tnw.set_message('test timeout', 1000);
        """)
    wait_for_selector(notebook.browser, "#notification_test", visible=True)

    assert widget_message(notebook, "test") == "test timeout", "timeout: message is correct"
    wait_for_selector(notebook.browser, "#notification_test", obscures=True)
    assert widget_message(notebook, "test") == "", "timeout: message was cleared"

    # test click callback
    notebook.browser.execute_script("""
        var tnw = IPython.notification_area.widget('test');
        tnw._clicked = false;
        tnw.set_message('test click', undefined, function () {
            tnw._clicked = true;
            return true;
        });
        """)
    wait_for_selector(notebook.browser, "#notification_test", visible=True)

    assert widget_message(notebook, "test") == "test click", "callback: message is correct"

    notebook.browser.find_element_by_id("notification_test").click()
    wait_for_script_to_return_true(notebook.browser,
                                   'return IPython.notification_area.widget("test")._clicked;')
    wait_for_selector(notebook.browser, "#notification_test", obscures=True)

    assert widget_message(notebook, "test") == "", "callback: message was cleared"



"""Tornado handler for bundling notebooks."""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from ipython_genutils.importstring import import_item
from tornado import web, gen

from notebook.utils import maybe_future, url2path
from notebook.base.handlers import IPythonHandler
from notebook.services.config import ConfigManager

from . import tools


class BundlerHandler(IPythonHandler):
    def initialize(self):
        """Make tools module available on the handler instance for compatibility
        with existing bundler API and ease of reference."""
        self.tools = tools
    
    def get_bundler(self, bundler_id):
        """
        Get bundler metadata from config given a bundler ID.
        
        Parameters
        ----------
        bundler_id: str
            Unique bundler ID within the notebook/bundlerextensions config section
        
        Returns
        -------
        dict
            Bundler metadata with label, group, and module_name attributes
        
        
        Raises
        ------
        KeyError
            If the bundler ID is unknown
        """
        cm = ConfigManager()
        return cm.get('notebook').get('bundlerextensions', {})[bundler_id]

    @web.authenticated
    @gen.coroutine
    def get(self, path):
        """Bundle the given notebook.
        
        Parameters
        ----------
        path: str
            Path to the notebook (path parameter)
        bundler: str
            Bundler ID to use (query parameter)
        """
        bundler_id = self.get_query_argument('bundler')
        model = self.contents_manager.get(path=url2path(path))

        try:
            bundler = self.get_bundler(bundler_id)
        except KeyError as e:
            raise web.HTTPError(400, 'Bundler %s not enabled' %
                                bundler_id) from e

        module_name = bundler['module_name']
        try:
            # no-op in python3, decode error in python2
            module_name = str(module_name)
        except UnicodeEncodeError:
            # Encode unicode as utf-8 in python2 else import_item fails
            module_name = module_name.encode('utf-8')
        
        try:
            bundler_mod = import_item(module_name)
        except ImportError as e:
            raise web.HTTPError(500, 'Could not import bundler %s ' %
                                bundler_id) from e

        # Let the bundler respond in any way it sees fit and assume it will
        # finish the request
        yield maybe_future(bundler_mod.bundle(self, model))

_bundler_id_regex = r'(?P<bundler_id>[A-Za-z0-9_]+)'

default_handlers = [
    (r"/bundle/(.*)", BundlerHandler)
]
